/*
by Claude Heiland-Allen 2022
see <https://mathr.co.uk/blog/2022-06-24_counting_artificial_neural_networks.html>
usage: gcc -O3 -fopenmp -std=c99 -Wall -Wextra -pedantic pruning.c -o pruning && ./pruning [limit [bias]]
note: -fopenmp for parallelism is optional
note: -std=c99 may be replaced with -std=c23 if the compiler supports it
*/

#include <inttypes.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>

#ifndef BITS
#error BITS not defined
#endif

#if BITS == 64
typedef uint64_t UINT;
#define SCAN SCNu64
#define PRINT PRIu64
#else
#if BITS == 32
typedef uint32_t UINT;
#define SCAN SCNu32
#define PRINT PRIu32
#else
#if BITS == 16
typedef uint16_t UINT;
#define SCAN SCNu16
#define PRINT PRIu16
#else
#if BITS == 8
typedef uint8_t UINT;
#define SCAN SCNu8
#define PRINT PRIu8
#else
#error BITS not defined to a supported value (8, 16, 32, 64)
#endif
#endif
#endif
#endif

#if __has_include("stdckdint.h")
/* C23 */
#include <stdckdint.h>
#else
/* gcc 5+, clang 3.8+ */
#define ckd_add(result,a,b) __builtin_add_overflow((a), (b), (result))
#define ckd_mul(result,a,b) __builtin_mul_overflow((a), (b), (result))
#endif

int main(int argc, char **argv)
{
  /*
    A pessimistic bound is sum <= (p + 1) * (p + bias) * p.
    Assuming bias is 0 or 1, then (p+1)^3 is bigger than that bound,
    So a pessimistic bound is p <= floor(cbrt(UINTMAX) - 1).
    If that holds, overflow should never occur.
    Due to pruning, overflow might not occur even for higher values.
    For example with 8 bits, the limit is calculated as 5 terms,
    but 11 can be calculated safely (without overflow).
    However, with 16 bits, the limit is calculated as 40 terms,
    but only 23 can be calculated with 16 bit accumulator, as term 24
    is 75268 > 2^16-1.
    For this reason, the count is always 64 bits unsigned.
  */
  UINT limit = floor(cbrt((UINT)~(UINT)0) - 1);
  if (argc > 1)
  {
    UINT user_limit = limit;
    if (1 != sscanf(argv[1], "%" SCAN, &user_limit))
    {
      fprintf(stderr, "ERROR: could not parse limit setting\n");
      return 1;
    }
    if (user_limit > limit)
    {
      fprintf(stderr, "WARNING: high limit setting might lead to overflow\n");
    }
    /* let user specify 0 to use default */
    if (user_limit > 0)
    {
      limit = user_limit;
    }
  }
  UINT bias = 0;
  if (argc > 2)
  {
    UINT user_bias = bias;
    if (1 != sscanf(argv[2], "%" SCAN, &user_bias))
    {
      fprintf(stderr, "ERROR: could not parse bias setting\n");
      return 1;
    }
    if (user_bias > 1)
    {
      fprintf(stderr, "WARNING: high bias setting might lead to overflow\n");
    }
    bias = user_bias;
  }
  #pragma omp parallel for ordered schedule(dynamic,1)
  for (UINT p = 1; p <= limit; ++p)
  {
    uint64_t count = 0;
    for (UINT m = 0; m <= p; ++m)
    {
      UINT s[m + 2];
      for (UINT i = 0; i < m + 1; ++i)
      {
        s[i] = 1;
      }
      s[m + 1] = 0; /* pre-incremented by loop before first sum */
      while (1)
      {
        UINT carry = 1;
        for (UINT j = m + 2; carry && j > 0; --j)
        {
          if (ckd_add(&s[j - 1], s[j - 1], carry))
          {
            fprintf(stderr, "ERROR: overflow in carry\n");
            fprintf(stderr, "j = %" PRINT ", s[j - 1] = %" PRINT ", carry = %" PRINT "\n", j, s[j - 1], carry);
            abort();
          }
          if (s[j - 1] > p)
          {
            s[j - 1] = 1;
            carry = 1;
          }
          else
          {
            carry = 0;
          }
        }
        if (carry)
        {
          break;
        }
        UINT sum = 0;
        UINT prefix = 0;
        for (UINT j = 0; sum <= p && j < m + 2 - 1; ++j)
        {
          UINT biased;
          if (ckd_add(&biased, s[j], bias))
          {
            fprintf(stderr, "ERROR: overflow adding bias\n");
            fprintf(stderr, "j = %" PRINT ", s[j] = %" PRINT ", biased = %" PRINT "\n", j, s[j], biased);
            abort();
          }
          UINT product;
          if (ckd_mul(&product, biased, s[j + 1]))
          {
            fprintf(stderr, "ERROR: overflow in product\n");
            fprintf(stderr, "j = %" PRINT ", s[j] + bias = %" PRINT ", s[j + 1] = %" PRINT ", product = %" PRINT "\n", j, biased, s[j + 1], product);
            abort();
          }
          if (ckd_add(&sum, sum, product))
          {
            fprintf(stderr, "ERROR: overflow in sum\n");
            fprintf(stderr, "j = %" PRINT ", sum = %" PRINT ", product = %" PRINT "\n", j, sum, product);
            abort();
          }
          prefix = j;
        }
        if (sum == p)
        {
          if (ckd_add(&count, count, 1))
          {
            fprintf(stderr, "ERROR: overflow in count\n");
            abort();
          }
        }
        else if (sum > p)
        {
          /*
            whatever the suffix, the sum will be too large
            skip to end of range by setting all to p
          */
          for (UINT j = prefix + 1; j < m + 2; ++j)
          {
            s[j] = p;
          }
        }
      }
    }
    #pragma omp ordered
    {
      printf("%" PRIu64 "\n", count);
      fflush(stdout);
    }
  }
  return 0;
}
